/* $Id: atari_dc.c,v 1.32 2002/05/17 00:16:52 chris Exp $
 *
 * SEGA Dreamcast support using KallistiOS (http://cadcdev.sourceforge.net)
 * (c) 2002 Christian Groessler (chris@groessler.org)
 */

#include <stdarg.h>
#include <stdlib.h>
#include <kos.h>
#include <sys/types.h>
#include "atari.h"
#include "config.h"
#include "sio.h"
#include "binload.h"
#include "input.h"
#include "rt-config.h"
#include "colours.h"
#include "ui.h"
#include "time.h"
#include "kb_ui.h"
#ifdef SOUND
#include <dc/g2bus.h>
#include <arm/aica_cmd_iface.h>
#include "sound.h"
#define _TYPEDEF_H   /* to prevent pokeysnd.h to create uint32 type #defines */
#include "pokeysnd.h"
#endif

#define REPEAT_DELAY 2000
#define REPEAT_INI_DELAY (3 * REPEAT_DELAY)
/*#define DEBUG*/
/*#define DOUBLE_BUFFER*/
/*#define HZ_TEST*/
/*#define MY_RELEASE*/

int stat(const char *, struct stat *);
static void calc_palette(void);
static int check_tray_open(void);
static void autostart(void);
static void controller_update(void);
#ifndef DOUBLE_BUFFER
static void dc_atari_sync(void);
#endif
#ifdef SOUND
static void dc_sound_init(void);
#endif

static uint8 mcont[4];
static cont_cond_t mcond[4];
static uint8 mkb;
static int num_cont;   /* # of controllers found */
static int su_first_call = TRUE;
static int b_ui_leave = FALSE;
int emulate_paddles = FALSE;
int glob_snd_ena = TRUE;

extern char curr_disk_dir[];
extern char curr_cart_dir[];
extern char curr_exe_dir[];

#define dsprate 22050

#if 0
static vid_mode_t mymode = {
	320, 240,
	VID_PIXELDOUBLE|VID_LINEDOUBLE,
	CT_ANY,
	PM_RGB565,
	262, 857,
	0xA4, 0x18,
	/*0x15, 0x104,*/
	0x15, 261,
	141, 843,
	24, 263*2,
	0, 1,
	{ 0, 0, 0, 0 }
};
#endif

void Atari_Initialise(int *argc, char *argv[])
{
	static int fc = TRUE;

	if (! fc) return; else fc = FALSE;

#ifndef DEBUG
	/* Bother us with output only if something died */
	dbglog_set_level(DBG_DEAD);
#endif
	
	/* Set the video mode */
	vid_set_mode(DM_320x240, PM_RGB565);
#if 0
	mymode.cable_type = vid_check_cable();
	vid_set_mode_ex(&mymode);
#endif
	calc_palette();
}

#ifdef SUPPORTS_ATARI_CONFIGINIT

void Atari_ConfigInit(void)  /* not currently used */
{
	strcpy(atari_osa_filename, "/cd/atariosa.rom");
	strcpy(atari_osb_filename, "/cd/atariosb.rom");
	strcpy(atari_xlxe_filename, "/cd/atarixl.rom");
	strcpy(atari_basic_filename, "/cd/ataribas.rom");
	strcpy(atari_5200_filename, "/cd/5200.rom");
	strcpy(atari_disk_dirs[0], "/cd/disks");
	disk_directories = 1;
	strcpy(atari_rom_dir, "/cd/cartridges");
	strcpy(atari_h1_dir, "/cd");
	strcpy(atari_h2_dir, "/cd");
	strcpy(atari_h3_dir, "/cd");
	strcpy(atari_h4_dir, "/cd");
	strcpy(atari_exe_dir, "/cd/files");
	strcpy(atari_state_dir, "/vmu");
}

#endif

int Atari_Exit(int run_monitor)
{
	arch_reboot();
	return(0);  /* not reached */
}

static void controller_update(void)
{
	int i;
	for (i=0; i<num_cont; i++) {
		if (cont_get_cond(mcont[i], &mcond[i])) {
#ifdef DEBUG
			printf("controller_update: error getting controller status\n");
#endif
		}
	}
}

static int consol_keys(void)
{
	cont_cond_t *cond = &mcond[0];

	if (! (cond->buttons & CONT_START)) {
		if (ui_is_active)
			arch_reboot();
		else {
			if (! (cond->buttons & CONT_X))
				return(AKEY_COLDSTART);
			else
				return(AKEY_WARMSTART);
		}
	}

	if (machine_type != MACHINE_5200) {
		if (! (cond->buttons & CONT_B)) {
			if (! b_ui_leave) {
				key_consol &= ~CONSOL_START;
			}
			else {
				key_consol |= CONSOL_START;
			}
		}
		else {
			b_ui_leave = FALSE;
			key_consol |= CONSOL_START;
		}

		if (! (cond->buttons & CONT_Y)) {
			key_consol &= ~CONSOL_SELECT;
		}
		else {
			key_consol |= CONSOL_SELECT;
		}

		if (! (cond->buttons & CONT_X)) {
			key_consol &= ~CONSOL_OPTION;
		}
		else {
			key_consol |= CONSOL_OPTION;
		}
	}
	else {
		if (! (cond->buttons & CONT_X)) {  /* 2nd action button */
			key_shift = 1;
		}
		else {
			key_shift = 0;
		}
		if (! ui_is_active) {
			if (! (cond->buttons & CONT_B)) {  /* testtest */
				if (! b_ui_leave) {
					return(AKEY_4);   /* ??? at least on the dc kb "4" starts some games?? */
				}
			}
			else {
				b_ui_leave = FALSE;
			}
			if (! (cond->buttons & CONT_Y)) {  /* testtest */
				return(AKEY_5200_START);
			}
		}
	}
	return(AKEY_NONE);
}

/*
 * do some basic keyboard emulation for the controller,
 * so that the emulator menu can be used without keyboard
 * (basically for selecting bin files/disk images)
 */
static int controller_kb(void)
{
	static int prev_up = FALSE, prev_down = FALSE, prev_a = FALSE,
		prev_r = FALSE, prev_left = FALSE, prev_right = FALSE,
		prev_b = FALSE;
	static int repdelay = REPEAT_DELAY;
	cont_cond_t *cond = &mcond[0];

	if (! num_cont) return(AKEY_NONE);  /* no controller present */

	repdelay--;
	if (repdelay < 0) repdelay = REPEAT_DELAY;

	if (cond->ltrig > 250) {
		return(AKEY_UI);
	}
#ifdef KB_UI
	if (! ui_is_active && cond->rtrig > 250) {
		controller_update();
		return(AKEY_KEYB);
	}
#endif

	if (ui_is_active) {
		if (! (cond->buttons & CONT_DPAD_UP)) {
			prev_down = FALSE;
			if (! prev_up) {
				repdelay = REPEAT_INI_DELAY;
				prev_up = 1;
				return(AKEY_UP);
			}
			else {
				if (! repdelay) {
					return(AKEY_UP);
				}
			}
		}
		else {
			prev_up = FALSE;
		}

		if (! (cond->buttons & CONT_DPAD_DOWN)) {
			prev_up = FALSE;
			if (! prev_down) {
				repdelay = REPEAT_INI_DELAY;
				prev_down = TRUE;
				return(AKEY_DOWN);
			}
			else {
				if (! repdelay) {
					return(AKEY_DOWN);
				}
			}
		}
		else {
			prev_down = FALSE;
		}

		if (! (cond->buttons & CONT_DPAD_LEFT)) {
			prev_right = FALSE;
			if (! prev_left) {
				repdelay = REPEAT_INI_DELAY;
				prev_left = TRUE;
				return(AKEY_LEFT);
			}
			else {
				if (! repdelay) {
					return(AKEY_LEFT);
				}
			}
		}
		else {
			prev_left = FALSE;
		}

		if (! (cond->buttons & CONT_DPAD_RIGHT)) {
			prev_left = FALSE;
			if (! prev_right) {
				repdelay = REPEAT_INI_DELAY;
				prev_right = TRUE;
				return(AKEY_RIGHT);
			}
			else {
				if (! repdelay) {
					return(AKEY_RIGHT);
				}
			}
		}
		else {
			prev_right = FALSE;
		}

		if (! (cond->buttons & CONT_A)) {
			if (! prev_a) {
				prev_a = TRUE;
				return(AKEY_RETURN);
			}
		}
		else {
			prev_a = FALSE;
		}

		if (! (cond->buttons & CONT_B)) {
			if (! prev_b) {
				prev_b = TRUE;
				b_ui_leave = TRUE;   /* B must be released again */
				return(AKEY_ESCAPE);
			}
		}
		else {
			prev_b = FALSE;
		}

		if (cond->rtrig > 250) {
			if (! prev_r) {
				prev_r = TRUE;
				return(AKEY_ESCAPE);
			}
		}
		else {
			prev_r = FALSE;
		}
	}
	return(AKEY_NONE);
}

int Atari_Keyboard(void)
{
	static int old_open_tray = FALSE;
	int open_tray;
	int keycode;

	open_tray = check_tray_open();
	if (open_tray != old_open_tray) {
		old_open_tray = open_tray;
		if (open_tray == TRUE) {
			return(AKEY_UI);
		}
		else {
			cdrom_reinit();
			iso_ioctl(0, NULL, 0);
			strcpy(curr_disk_dir, "/");
			strcpy(curr_cart_dir, "/");
			strcpy(curr_exe_dir, "/");
		}
	}

	if (ui_is_active) controller_update();

	if (num_cont && (keycode = consol_keys()) != AKEY_NONE) return(keycode);

#ifdef KB_UI
	if (inject_key != AKEY_NONE) {
		keycode = inject_key;
		inject_key = AKEY_NONE;
		return(keycode);
	}
#endif

	keycode = controller_kb();
	if (keycode != AKEY_NONE) return(keycode);

	keycode = kbd_get_key();
	if (keycode == -1) return(AKEY_NONE);

#ifdef DEBUG
	printf("DC key: %04X (%c)\n", keycode, keycode > 32 && keycode < 127 ? keycode : '.');
#endif

	switch (keycode) {

	/* OPTION / SELECT / START keys */
	/*key_consol = CONSOL_NONE;  -- already set in consol_keys... */
	case 0x3b00:
		key_consol &= (~CONSOL_OPTION);
		return(AKEY_NONE);
	case 0x3c00:
		key_consol &= (~CONSOL_SELECT);
		return(AKEY_NONE);
	case 0x3d00:
		key_consol &= (~CONSOL_START);
		return(AKEY_NONE);

	case 0x1b:  /* ESC */
		keycode = AKEY_ESCAPE;
		break;
	case '`':
		keycode = AKEY_CAPSTOGGLE;
		break;
	case '!':
		keycode = AKEY_EXCLAMATION;
		break;
	case '"':
		keycode = AKEY_DBLQUOTE;
		break;
	case '#':
		keycode = AKEY_HASH;
		break;
	case '$':
		keycode = AKEY_DOLLAR;
		break;
	case '%':
		keycode = AKEY_PERCENT;
		break;
	case '&':
		keycode = AKEY_AMPERSAND;
		break;
	case '\'':
		keycode = AKEY_QUOTE;
		break;
	case '@':
		keycode = AKEY_AT;
		break;
	case '(':
		keycode = AKEY_PARENLEFT;
		break;
	case ')':
		keycode = AKEY_PARENRIGHT;
		break;
	case '[':
		keycode = AKEY_BRACKETLEFT;
		break;
	case ']':
		keycode = AKEY_BRACKETRIGHT;
		break;
	case '<':
		keycode = AKEY_LESS;
		break;
	case '>':
		keycode = AKEY_GREATER;
		break;
	case '=':
		keycode = AKEY_EQUAL;
		break;
	case '?':
		keycode = AKEY_QUESTION;
		break;
	case '-':
		keycode = AKEY_MINUS;
		break;
	case '+':
		keycode = AKEY_PLUS;
		break;
	case '*':
		keycode = AKEY_ASTERISK;
		break;
	case '/':
		keycode = AKEY_SLASH;
		break;
	case ':':
		keycode = AKEY_COLON;
		break;
	case ';':
		keycode = AKEY_SEMICOLON;
		break;
	case ',':
		keycode = AKEY_COMMA;
		break;
	case '.':
		keycode = AKEY_FULLSTOP;
		break;
	case '_':
		keycode = AKEY_UNDERSCORE;
		break;
	case '^':
		keycode = AKEY_CIRCUMFLEX;
		break;
	case '\\':
		keycode = AKEY_BACKSLASH;
		break;
	case '|':
		keycode = AKEY_BAR;
		break;
	case ' ':
		keycode = AKEY_SPACE;
		break;
	case '0':
		keycode = AKEY_0;
		break;
	case '1':
		keycode = AKEY_1;
		break;
	case '2':
		keycode = AKEY_2;
		break;
	case '3':
		keycode = AKEY_3;
		break;
	case '4':
		keycode = AKEY_4;
		break;
	case '5':
		keycode = AKEY_5;
		break;
	case '6':
		keycode = AKEY_6;
		break;
	case '7':
		keycode = AKEY_7;
		break;
	case '8':
		keycode = AKEY_8;
		break;
	case '9':
		keycode = AKEY_9;
		break;
	case 'a':
		keycode = AKEY_a;
		break;
	case 'b':
		keycode = AKEY_b;
		break;
	case 'c':
		keycode = AKEY_c;
		break;
	case 'd':
		keycode = AKEY_d;
		break;
	case 'e':
		keycode = AKEY_e;
		break;
	case 'f':
		keycode = AKEY_f;
		break;
	case 'g':
		keycode = AKEY_g;
		break;
	case 'h':
		keycode = AKEY_h;
		break;
	case 'i':
		keycode = AKEY_i;
		break;
	case 'j':
		keycode = AKEY_j;
		break;
	case 'k':
		keycode = AKEY_k;
		break;
	case 'l':
		keycode = AKEY_l;
		break;
	case 'm':
		keycode = AKEY_m;
		break;
	case 'n':
		keycode = AKEY_n;
		break;
	case 'o':
		keycode = AKEY_o;
		break;
	case 'p':
		keycode = AKEY_p;
		break;
	case 'q':
		keycode = AKEY_q;
		break;
	case 'r':
		keycode = AKEY_r;
		break;
	case 's':
		keycode = AKEY_s;
		break;
	case 't':
		keycode = AKEY_t;
		break;
	case 'u':
		keycode = AKEY_u;
		break;
	case 'v':
		keycode = AKEY_v;
		break;
	case 'w':
		keycode = AKEY_w;
		break;
	case 'x':
		keycode = AKEY_x;
		break;
	case 'y':
		keycode = AKEY_y;
		break;
	case 'z':
		keycode = AKEY_z;
		break;
	case 'A':
		keycode = AKEY_A;
		break;
	case 'B':
		keycode = AKEY_B;
		break;
	case 'C':
		keycode = AKEY_C;
		break;
	case 'D':
		keycode = AKEY_D;
		break;
	case 'E':
		keycode = AKEY_E;
		break;
	case 'F':
		keycode = AKEY_F;
		break;
	case 'G':
		keycode = AKEY_G;
		break;
	case 'H':
		keycode = AKEY_H;
		break;
	case 'I':
		keycode = AKEY_I;
		break;
	case 'J':
		keycode = AKEY_J;
		break;
	case 'K':
		keycode = AKEY_K;
		break;
	case 'L':
		keycode = AKEY_L;
		break;
	case 'M':
		keycode = AKEY_M;
		break;
	case 'N':
		keycode = AKEY_N;
		break;
	case 'O':
		keycode = AKEY_O;
		break;
	case 'P':
		keycode = AKEY_P;
		break;
	case 'Q':
		keycode = AKEY_Q;
		break;
	case 'R':
		keycode = AKEY_R;
		break;
	case 'S':
		keycode = AKEY_S;
		break;
	case 'T':
		keycode = AKEY_T;
		break;
	case 'U':
		keycode = AKEY_U;
		break;
	case 'V':
		keycode = AKEY_V;
		break;
	case 'W':
		keycode = AKEY_W;
		break;
	case 'X':
		keycode = AKEY_X;
		break;
	case 'Y':
		keycode = AKEY_Y;
		break;
	case 'Z':
		keycode = AKEY_Z;
		break;

	case 0x3a00:  /* F1 */
		keycode = AKEY_UI;
		break;

	case 0x3e00:  /* F5 */
		keycode = AKEY_COLDSTART;
		break;

	case 0x4500:  /* F12 */
		arch_reboot();

	/* cursor keys */
	case 0x5200:
		keycode = AKEY_UP;
		break;
	case 0x5100:
		keycode = AKEY_DOWN;
		break;
	case 0x5000:
		keycode = AKEY_LEFT;
		break;
	case 0x4f00:
		keycode = AKEY_RIGHT;
		break;

	case 0x6500:  /* S3 */
		keycode = AKEY_ATARI;
		break;

	case 0x4a00:  /* Home */
		keycode = AKEY_CLEAR;
		break;
	case 0x4d00:  /* End */
		keycode = AKEY_HELP;
		break;
	case 0x4800:  /* Break/Pause */
		keycode = AKEY_BREAK;
		break;

	case 0x4c00:  /* Del */
		if (key_shift)
			keycode = AKEY_DELETE_LINE;
		else
			keycode |= AKEY_DELETE_CHAR;
		break;
	case 0x4900:  /* Ins */
		if (key_shift)
			keycode = AKEY_INSERT_LINE;
		else
			keycode |= AKEY_INSERT_CHAR;
		break;

	case 127:
	case 8:
		keycode = AKEY_BACKSPACE;
		break;
	case 13:
	case 10:
		keycode = AKEY_RETURN;
		break;
	default:
		keycode = AKEY_NONE;
		break;
	}
	return keycode;
}

int Atari_PORT(int num)
{
	cont_cond_t *cond;
	int retval = 0xff;
	int cix = 0;

	if (num) {
		if (machine_type != MACHINE_OSA && machine_type != MACHINE_OSB) return(retval);
		cix = 2;   /* js #2 and #3 */
	}

	if (num_cont < 1 + cix) return(retval);  /* no controller present */

	cond = &mcond[cix];
	if (! emulate_paddles) {
		if (! (cond->buttons & CONT_DPAD_UP)) {
			retval &= 0xfe;
		}
		if (! (cond->buttons & CONT_DPAD_DOWN)) {
			retval &= 0xfd;
		}
		if (! (cond->buttons & CONT_DPAD_LEFT)) {
			retval &= 0xfb;
		}
		if (! (cond->buttons & CONT_DPAD_RIGHT)) {
			retval &= 0xf7;
		}
                /* if joypad not used, try the joystick instead */
		if (retval == 0xff) {
			if (cond->joyx > 127 + 64) {  /* right */
				retval &= 0xf7;
			}
			if (cond->joyx < 127 - 64) {  /* left */
				retval &= 0xfb;
			}
			if (cond->joyy > 127 + 64) {  /* down */
				retval &= 0xfd;
			}
			if (cond->joyy < 127 - 64) {  /* up */
				retval &= 0xfe;
			}
		}
		if (num_cont > 1 + cix) {
			cond = &mcond[cix+1];
			if (! (cond->buttons & CONT_DPAD_UP)) {
				retval &= 0xef;
			}
			if (! (cond->buttons & CONT_DPAD_DOWN)) {
				retval &= 0xdf;
			}
			if (! (cond->buttons & CONT_DPAD_LEFT)) {
				retval &= 0xbf;
			}
			if (! (cond->buttons & CONT_DPAD_RIGHT)) {
				retval &= 0x7f;
			}
			/* if joypad not used, try the joystick instead */
			if ((retval & 0xf0) == 0xf0) {
				if (cond->joyx > 127 + 64) {  /* right */
					retval &= 0x7f;
				}
				if (cond->joyx < 127 - 64) {  /* left */
					retval &= 0xbf;
				}
				if (cond->joyy > 127 + 64) {  /* down */
					retval &= 0xdf;
				}
				if (cond->joyy < 127 - 64) {  /* up */
					retval &= 0xef;
				}
			}
		}
	}
	else {  /* emulate paddles */
		/* 1st paddle trigger */
		if (! (cond->buttons & CONT_A)) {
			retval &= (cix > 1) ? 0xbf : 0xfb;
		}
		if (num_cont > 1 + cix) {
			cond = &mcond[cix+1];

			/* 2nd paddle trigger */
			if (! (cond->buttons & CONT_A)) {
				retval &= (cix > 1) ? 0x7f : 0xf7;
			}
		}
	}

	return(retval);
}

int Atari_POT(int num)
{
	int val;
	cont_cond_t *cond;

	if (machine_type != MACHINE_5200) {
		if (emulate_paddles) {
			if (num + 1 > num_cont) return(228);
			if (machine_type != MACHINE_OSA && machine_type != MACHINE_OSB && num > 3) return(228);

			cond = &mcond[num];
			val = cond->joyx;
			val = val * 228 / 255;
			if (val > 227) return(1);
			return(228 - val);
		}
		else {
			return(228);
		}
	}
	else {  /* 5200 version:
		 *
		 * num / 2: which controller
		 * num & 1 == 0: x axis
		 * num & 1 == 1: y axis
		 */

		if (num / 2 + 1 > num_cont) return(joy_5200_center);

		cond = &mcond[num / 2];
		val = (num & 1) ? cond->joyy : cond->joyx;

		/* normalize into 5200 range */
		/* @@@ is this ok? it's already late now @@@ */
		if (val == 127) return(joy_5200_center);
		if (val < 127) {
			/*val -= joy_5200_min;*/
			val = val * (joy_5200_center - joy_5200_min) / 127;
			return(val + joy_5200_min);
		}
		else {
			val = val * joy_5200_max / 255;
			if (val < joy_5200_center)
				val = joy_5200_center;
			return(val);
		}
	}
}

int Atari_TRIG(int num)
{
	cont_cond_t *cond;

	if (num + 1 > num_cont) return(1);  /* no controller present */
	if (machine_type != MACHINE_OSA && machine_type != MACHINE_OSB && num > 1) return(1);

	cond = &mcond[num];
	if (! (cond->buttons & CONT_A)) {
		return(0);
	}

	return(1);
}

static uint16 mypal[256];

static void calc_palette(void)
{
	int i;
	int rgb;
	int red;
	int green;
	int blue;
	uint16 pv;

	for (i = 0; i < 256; i++) {
		rgb = colortable[i];

		red = (rgb & 0x00ff0000) >> 16;
		green = (rgb & 0x0000ff00) >> 8;
		blue = (rgb & 0x000000ff) ;

		pv = ((red >> 3) << 11) |
			((green >> 2) << 5) |
			(blue >> 3);

		mypal[i] = pv;
	}
}

#ifdef DOUBLE_BUFFER

static void vbl_wait(void)
{
	static int last_cntr = 0;

	while (maple_state.vbl_cntr == last_cntr) ;
	last_cntr = maple_state.vbl_cntr;
}

#endif /* #ifdef DOUBLE_BUFFER */

#ifdef DIRTYRECT
#ifdef DOUBLE_BUFFER

static UBYTE old_sd[2][ATARI_HEIGHT * ATARI_WIDTH/8];

void Atari_DisplayScreen(UBYTE *screen)
{
	static unsigned int sbase = 0;  /* screen base of current frame */
	unsigned int x, y, m, j;
	uint16 *vram;
	UBYTE *osd, *nsd;  /* old screen-dirty, new screen-dirty */

	/*vid_border_color(0, 0, 0);*/
	if (sbase) {
		sbase = 0;
		osd = old_sd[0];
		nsd = old_sd[1];
	}
	else {
		sbase = 1024 * 768 * 4;
		osd = old_sd[1];
		nsd = old_sd[0];
	}

	vram = (uint16*)(0xA5000000 + sbase) + (vid_mode->height - 240) / 2 * vid_mode->width;
	vram += (vid_mode->height - 240) / 2 * vid_mode->width;
	vram += (vid_mode->width - 320) / 2 - (ATARI_WIDTH - 320) / 2 ;

	for (m=0, x=0, y=0; m < ATARI_HEIGHT * ATARI_WIDTH / 8; m++) {
		nsd[m] = screen_dirty[m];  /* remember for other page */
		if (/*1 ||*/ screen_dirty[m] || osd[m]) {
			if (x - (ATARI_WIDTH - 320) / 2 <= 320 - 8) {
				/* Draw eight pixels (x,y), (x+1,y),...,(x+7,y) here */
				for (j=0; j<8; j++) {
					*(vram + x + j) = mypal[*(screen + x + j)];
				}
				screen_dirty[m] = 0;
			}
		}
		x += 8;
		if (x >= ATARI_WIDTH) {
			x = 0;
			y++;
			vram += vid_mode->width;
			screen += ATARI_WIDTH;
		}
	}
	vbl_wait();
	/*vid_border_color(255, 255, 255);*/
	vid_set_start(sbase);
}

#else

void Atari_DisplayScreen(UBYTE *screen)
{
	unsigned int x, y, m, j;
	uint16 *vram = vram_s;

	/* point to current line of atari screen in display memory */
	vram += (vid_mode->height - 240) / 2 * vid_mode->width;

	/* adjust left column to screen resolution */
	vram += (vid_mode->width - 320) / 2 - (ATARI_WIDTH - 320) / 2 ;

	for (m=0, x=0, y=0; m < ATARI_HEIGHT * ATARI_WIDTH / 8; m++) {
		if (screen_dirty[m]) {
			if (x - (ATARI_WIDTH - 320) / 2 <= 320 - 8) {
				/* Draw eight pixels (x,y), (x+1,y),...,(x+7,y) here */
				for (j=0; j<8; j++) {
					*(vram + x + j) = mypal[*(screen + x + j)];
				}
				screen_dirty[m] = 0;
			}
		}
		x += 8;
		if (x >= ATARI_WIDTH) {
			x = 0;
			y++;
			vram += vid_mode->width;
			screen += ATARI_WIDTH;
		}
	}
}

#endif /* not define DOUBLE_BUFFER */
#else /* above DIRTY_RECT, below not */
#ifdef DOUBLE_BUFFER

void Atari_DisplayScreen(UBYTE *screen)
{
	static unsigned int sbase = 0;  /* screen base of current frame */
	int i, j;
	uint16 *vram;

	if (sbase) {
		sbase = 0;
	}
	else {
		sbase = 1024 * 768 * 4;
	}

	vram = (uint16*)(0xA5000000|sbase) + (vid_mode->height - 240) / 2 * vid_mode->width;

	for (i=0; i<240; i++) {         /* y */
		screen += (ATARI_WIDTH - 320) / 2;
		vram += (vid_mode->width - 320) / 2;
		for (j=0; j<320; j++) { /* x */
			*vram++ = mypal[*screen++];
		}
		vram += (vid_mode->width - 320) / 2;
		screen += (ATARI_WIDTH - 320) / 2;
	}
	vbl_wait();
	vid_set_start(sbase);
}

#else /* above DOUBLE_BUFFER, below not */

void Atari_DisplayScreen(UBYTE *screen)
{
	int i, j;
	uint16 *vram = vram_s;

	vram += (vid_mode->height - 240) / 2 * vid_mode->width;

	for (i=0; i<240; i++) {         /* y */
		screen += (ATARI_WIDTH - 320) / 2;
		vram += (vid_mode->width - 320) / 2;
		for (j=0; j<320; j++) { /* x */
			*vram++ = mypal[*screen++];
		}
		vram += (vid_mode->width - 320) / 2;
		screen += (ATARI_WIDTH - 320) / 2;
	}
}

#endif /* not defined DOUBLE_BUFFER */
#endif /* not defined DIRTY_RECT */

static void dc_controller_init(void)
{
	int p, u;

	for (p=0; p<MAPLE_PORT_COUNT; p++) {
		for (u=0; u<MAPLE_UNIT_COUNT; u++) {
			if (maple_device_func(p, u) & MAPLE_FUNC_CONTROLLER) {
				mcont[num_cont++] = maple_addr(p, u);
#ifdef DEBUG
				printf("dc_controller_init: found controller %d\n", mcont[num_cont-1]);
#endif
			}
		}
	}

	mkb = maple_first_kb();

#ifdef DEBUG
	printf("dc_controller_init: controller%sfound, keyboard%sfound\n", 
	       mcont[0] ? " " : " not ", mkb ? " " : " not ");
#endif
	return;
}

#if defined(MY_RELEASE) || defined(HZ_TEST)
static void cls(int xres, int yres)
{
	int x, y;
	for(x = 0; x < xres; x++)
		for(y = 0; y < yres; y++)
			vram_s[xres*y + x] = 0;
}

static void wait_a(void)
{
	cont_cond_t cond;

	do {
		if (cont_get_cond(mcont[0], &cond)) {
			return;
		}

	} while (cond.buttons & CONT_A);
}
#endif

extern tUIDriver dc_ui_driver;

KOS_INIT_FLAGS(INIT_IRQ);
KOS_INIT_ROMDISK(ROMDISK_NONE);

int main(int argc, char **argv)
{
	/* set own user interface variation */
	ui_driver = &dc_ui_driver;

	/* initialise Atari800 core */
	Atari800_Initialise(&argc, argv);

	/* initialize dc controllers */
	dc_controller_init();

#ifdef SOUND
	/* initialize sound */
	dc_sound_init();
#endif

	/*ser_console_init();*/

#ifdef MY_RELEASE
	cls(320,240);
	bfont_draw_str(vram_s+50*320+5, 320, 0, "Hallo Tobias und Dominik!");
 	bfont_draw_str(vram_s+57*320+5, 320, 0, "_________________________");
	wait_a();
	cls(320,240);
#endif

#ifdef DEBUG
        printf("\nFor Tobias & Dominik\n");
        printf("--------------------\n");
#endif

#ifdef HZ_TEST
	{
		uint32 s, ms, s2, ms2, z;
		char buffer[80];

		bfont_draw_str(vram_s+64*320+5, 320, 0, "chk dsply hz.. 500 vbls");
		timer_ms_gettime(&s, &ms);
		for (z=0; z<500; z++) {
			vbl_wait();
			/*vid_waitvbl();*/
		}
		timer_ms_gettime(&s2, &ms2);
		sprintf(buffer, "start time: %lu.%03lu\n", s, ms);
		bfont_draw_str(vram_s+89*320+5, 320, 0, buffer);
		sprintf(buffer, "end   time: %lu.%03lu\n", s2, ms2);
		bfont_draw_str(vram_s+114*320+5, 320, 0, buffer);
		sprintf(buffer, "diff(ms) = %ld\n", s2*1000 + ms2 - s * 1000 - ms);
		bfont_draw_str(vram_s+139*320+5, 320, 0, buffer);
		sprintf(buffer, "hz = %f\n", 500.0 / ((s2*1000 + ms2 - s * 1000 - ms) / 1000.0));
		bfont_draw_str(vram_s+164*320+5, 320, 0, buffer);
		wait_a();
		cls(320,240);
	}
#endif

	autostart();

	/* main loop */
	while(TRUE)
	{
		int keycode;

		keycode = Atari_Keyboard();

		switch (keycode) {
		case AKEY_5200_RESET:
			if (machine_type == MACHINE_5200) {
				Coldstart();
			}
			break;
		case AKEY_COLDSTART:
			Coldstart();
			break;
		case AKEY_WARMSTART:
			Warmstart();
			break;
		case AKEY_EXIT:
			Atari800_Exit(FALSE);
			arch_reboot();
#ifdef KB_UI
		case AKEY_KEYB:
			if (machine_type != MACHINE_5200) {
#ifdef SOUND
				Sound_Pause();
#endif
				kb_ui((UBYTE *)atari_screen);
				key_consol |= CONSOL_START;
				/*b_ui_leave = TRUE;  crashes when included!! why?? */
#ifdef SOUND
				Sound_Continue();
#endif
				controller_update();
			}
                        else {
#ifdef SOUND
				Sound_Pause();
#endif
				kb_ui_5200((UBYTE *)atari_screen);
#ifdef SOUND
				Sound_Continue();
#endif
				controller_update();
                        }
			break;
#endif /* #ifdef KB_UI */
		case AKEY_UI:
#ifdef SOUND
			Sound_Pause();
#endif
			ui((UBYTE *)atari_screen);
#ifdef SOUND
			Sound_Continue();
#endif
			controller_update();
			break;
		case AKEY_BREAK:
			key_break = 1;
			break;
		default:
			key_break = 0;
			key_code = keycode;
			break;
		}

		Atari800_Frame(EMULATE_FULL);
		Atari_DisplayScreen((UBYTE *)atari_screen);
#ifndef DOUBLE_BUFFER /* with double buffer we sync on vbl */
		dc_atari_sync();
		//vbl_wait();
#endif
		controller_update();  /* get new values from the controllers */
	}
}

/*
 * autostart: check for autorun.xxx files on the
 * CD root directory and start the first found
 */
static void autostart(void)
{
	struct stat sb;

	if (! stat("/cd/autorun.com", &sb)) {
		if (! (sb.st_mode & S_IFDIR)) {
			if (BIN_loader("/cd/autorun.com")) {
				Coldstart();
				return;
			}
		}
	}
	if (! stat("/cd/autorun.exe", &sb)) {
		if (! (sb.st_mode & S_IFDIR)) {
			if (BIN_loader("/cd/autorun.exe")) {
				Coldstart();
				return;
			}
		}
	}
	if (! stat("/cd/autorun.atr", &sb)) {
		if (! (sb.st_mode & S_IFDIR)) {
			SIO_Mount(1, "/cd/autorun.atr", TRUE);
			return;
		}
	}
}

int sscanf(const char *str, const char *format, ...)
{
	va_list args;
	int *res;

	if (strcmp(format, "%d")) return(0);   /* sanity check, we only support "%d" */

	va_start(args, format);
	res = va_arg(args, int *);
	*res = atoi(str);
        va_end(args);
	return(1);
}

static char mytmpnam[] = "tempname.oef";
char *tmpnam(char *space)
{
	if (space) {
		strcpy(space, mytmpnam);
		return(space);
	}
	else {
		return(mytmpnam);
	}
}
FILE *tmpfile(void) { return(NULL); }
int system(const char* path) { return 0; }

static struct tm localtime_tm;
struct tm *localtime(const time_t *clock)
{
	/* Get the time */
	localtime_r(clock, &localtime_tm);
	return(&localtime_tm);
}

static int odc;
DIR *opendir(const char *name)
{
	static file_t handle;
	handle = fs_open(name, O_DIR | O_RDONLY);
	if (! handle) return(NULL);
#ifndef NO_STAT
	if (strcmp(name, "/")) odc = 1;  /* no ".." in the root directory */
#endif
	return(&handle);
}

static struct dirent myd;

struct dirent *readdir(DIR *dirp)
{
#ifndef NO_STAT
	if (odc) {
		odc = 0;
		memset(&myd, 0, sizeof(struct dirent));
		strcpy(myd.d_name, "..");
		return(&myd);
	}
#endif
	return(fs_readdir(*dirp));
}

int closedir(DIR *dirp)
{
	fs_close(*dirp);
	return(0);
}

/* our super-duper stat */
/* only check, whether the S_IFDIR bit has to be set */
int stat(const char *path, struct stat *sb)
{
	file_t handle;

	memset(sb, 0, sizeof(struct stat));  /* preinitialize result */

#ifdef NO_STAT
#warning "DREAMCAST missing stat -- no directories"
	return(0);
#else

	/* special ".." case */
	if (strlen(path) > 3 &&
	    ! strcmp(path + strlen(path) - 3, "/..")) {
		sb->st_mode |= S_IFDIR;
		return(0);  /* success */
	}

	/* check if dir */
	if ((handle = fs_open(path, O_DIR | O_RDONLY))) {  /* is dir */
		sb->st_mode |= S_IFDIR;
	}
	else if ((handle = fs_open(path, O_RDONLY))) {  /* is file */
		/* nothing to do */
	}
	else {  /* is not here */
		return(-1);
	}
	fs_close(handle);
	return(0);  /* success */
#endif
}

int getchar(void) { return EOF; }
int unlink(const char *path) { return(fs_unlink(path)); }
int remove(const char *path) { return(unlink(path)); } 
void dc_perror(const char *x) { printf("perror: '%s'\n", x); }
void (*signal(int sig, void (*func)(int)))(int) { return (void *) 0; }

void SaveUBYTE( UBYTE *data, int num )
{
	return;
}
void SaveINT( int *data, int num ) 
{
	return;
}
void SaveUWORD( UWORD *data, int num )
{
	return;
}
void ReadUBYTE( UBYTE *data, int num )
{
	return;
}
void ReadINT( int *data, int num ) 
{
	return;
}
void ReadUWORD( UWORD *data, int num )
{
	return;
}
int SaveAtariState( char *filename, const char *mode, UBYTE SaveVerbose )
{
	return(FALSE);
}
int ReadAtariState( char *filename, const char *mode )
{
	return(FALSE);
}

#ifndef DOUBLE_BUFFER

static void dc_atari_sync(void)
{
	static unsigned long long nextclock = 0;
	unsigned long long curclock;
	uint32 s, ms;

	do {
		timer_ms_gettime(&s, &ms);
		curclock = ((unsigned long long)s << 32) + ms;
	} while (curclock < nextclock);
        
	nextclock = curclock + CLK_TCK / ((tv_mode == TV_PAL ? 50 : 60));
}

#endif /* #ifndef DOUBLE_BUFFER */
#ifdef SOUND

static int sound_enabled = TRUE;        /* sound: on or off */

#define FRAGSIZE       12		/* 1<<FRAGSIZE is size of sound buffer */

/* This will come from a seperately linked object file */
extern uint8 snd_stream_drv[];
extern uint8 snd_stream_drv_end[];

/* taken from KOS' snd_stream.c */
void dc_sound_init(void)
{
	unsigned int amt = snd_stream_drv_end - snd_stream_drv;

	spu_disable();
	spu_memset(0, 0, 0x31000);
	if (amt % 4)
		amt = (amt + 4) & ~3;
	spu_memload(0, snd_stream_drv, amt);
	spu_enable();
	timer_spin_sleep(10);
#ifdef STEREO
	Pokey_sound_init(FREQ_17_EXACT, dsprate, 2);
#else
	Pokey_sound_init(FREQ_17_EXACT, dsprate, 1);
#endif
}

void Sound_Pause(void)
{
	if (sound_enabled) {
		sound_enabled = FALSE;
		snd_stream_stop();
	}
}

void Sound_Continue(void)
{
	if (! sound_enabled && glob_snd_ena) {
		sound_enabled = TRUE;
		su_first_call = TRUE;
	}
}

void Sound_Exit(void)
{
#ifdef DEBUG
        printf("Sound_Exit called\n");
#endif
}

/* taken from KOS' snd_stream.c */
#define SPU_RAM_BASE		0xa0800000

/* taken from KOS' snd_stream.c */
static void kick_aica(void)
{
	g2_write_32(SPU_RAM_BASE + AICA_CMD, AICA_CMD_KICK);
	g2_fifo_wait();
	while (1) {
		int res;
		
		res=g2_read_32(SPU_RAM_BASE + AICA_CMD);
		g2_fifo_wait();
		if (!res)
			return;
	}
}

/* ripped from KOS */
unsigned int aica_getpos(int chan)
{
	aica_channel chanbuf;

	/* Get "real" buffer */
	chanbuf.cmd = AICA_CMD_UPDATE | AICA_UPDATE_GET_POS;
	spu_memload(AICA_CHANNEL(chan), &chanbuf, sizeof(chanbuf));
	kick_aica();
	spu_memread(&chanbuf, AICA_CHANNEL(chan), sizeof(chanbuf));
	return(chanbuf.pos);
}


/* taken from KOS' snd_stream.c */
void snd_stream_stop()
{
        aica_channel newval;
        
        /* stop stream */
        /* channel 0 */
        newval.cmd = AICA_CMD_STOP;
        spu_memload(AICA_CHANNEL(0), &newval, sizeof(newval));
#ifdef STEREO
        /* channel 1 */
        spu_memload(AICA_CHANNEL(1), &newval, sizeof(newval));
#endif
        kick_aica();
}


/* frag size and # values are from sound.c, not my invention :-) */

#define FRAG_BUFFER  0x11000   /* start of fragment buffer */
#define FRAG_BUFFER2 0x20000   /* start of fragment buffer for stereo 2nd channel */
#define FRAG_SIZE    0x80      /* size of one fragment */
#define FRAG_NUM     12        /* max. # of fragments in the buffer */
#define FRAG_BUFSZ   (FRAG_NUM * FRAG_SIZE)

static unsigned char fragbuf[FRAG_SIZE]; /* scratch buffer to generate sound data for a fragment */

/* some parts taken from KOS' snd_stream.c */
void Sound_Update(void)
{
        static int last_frag;
        int cur_pos, cur_frag, fill_frag, first_frag_to_fill, last_frag_to_fill;

	if (! sound_enabled) return;

        if (su_first_call) {
		aica_channel newval;

		su_first_call = FALSE;

                /* prefill buffers */
                Pokey_process(fragbuf, FRAG_SIZE);
		spu_memload(FRAG_BUFFER, fragbuf, FRAG_SIZE);
#if 0
		Pokey_process(fragbuf, FRAG_SIZE);
		spu_memload(FRAG_BUFFER + FRAG_SIZE, fragbuf, FRAG_SIZE);
		Pokey_process(fragbuf, FRAG_SIZE);
		spu_memload(FRAG_BUFFER + FRAG_SIZE * 2, fragbuf, FRAG_SIZE);
		Pokey_process(fragbuf, FRAG_SIZE);
		spu_memload(FRAG_BUFFER + FRAG_SIZE * 3, fragbuf, FRAG_SIZE);
		Pokey_process(fragbuf, FRAG_SIZE);
		spu_memload(FRAG_BUFFER + FRAG_SIZE * 4, fragbuf, FRAG_SIZE);
		Pokey_process(fragbuf, FRAG_SIZE);
		spu_memload(FRAG_BUFFER + FRAG_SIZE * 5, fragbuf, FRAG_SIZE);
#endif
		last_frag = 0 /*5*/;
#ifdef STEREO
#error STEREO not implemented yet!
#endif

                /* start streaming */
                /* use channel 0 */
                newval.cmd = AICA_CMD_START;
                newval.pos = FRAG_BUFFER;
                newval.type = SM_8BIT;
                newval.length = FRAG_BUFSZ;
                newval.loop = 1;
                newval.loopstart = 0;
                newval.loopend = FRAG_BUFSZ - 1;
                newval.freq = dsprate;
                newval.vol = 240;
                newval.pan = 127;
                spu_memload(AICA_CHANNEL(0), &newval, sizeof(newval));

                kick_aica();
        }

	/* get currently played position */
	cur_pos = aica_getpos(0);
	cur_frag = cur_pos / FRAG_SIZE;

	/* calc # of new fragments needed to fill sound buffer */

	first_frag_to_fill = last_frag + 1;
	if (first_frag_to_fill >= FRAG_NUM) first_frag_to_fill = 0;
	last_frag_to_fill = cur_frag;
	if (last_frag_to_fill < 0) last_frag_to_fill = FRAG_NUM - 1;

	fill_frag = first_frag_to_fill;
	while (fill_frag != last_frag_to_fill) {
		Pokey_process(fragbuf, FRAG_SIZE);
		spu_memload(FRAG_BUFFER + FRAG_SIZE * fill_frag, fragbuf, FRAG_SIZE);
		last_frag = fill_frag;
		fill_frag++;
		fill_frag %= FRAG_NUM;
		/*if (fill_frag >= FRAG_NUM) fill_frag = 0;*/
	}
}

#endif	/* #ifdef SOUND */

static int check_tray_open(void)
{
	int status, disk_type, retval;

	retval = cdrom_get_status(&status, &disk_type);
	if (retval == ERR_OK) {
		if ((status & 15) == 6) return(TRUE);
		return(FALSE);
	}
	else {
		return(FALSE); /* error case: assume tray not open */
	}
}
