/* $Id: ui_dc.c,v 1.5 2002/05/17 00:04:47 chris Exp $
 * Dreamcast port UI driver
 * (c) 2002  Christian Groessler
 * derived from wince/port/ui_wince.c by Vasyl Tsvirkunov
 */

#include <string.h>
#include <malloc.h>
#include "atari.h"
#include "ui.h"
#include "input.h"
#include "version.h"

#define CENTERED_MAIN_MENU  /* configurable, since I'm not decided yet, what looks better */

extern int emulate_paddles;
extern int glob_snd_ena;

static int DCUISelect(char* pTitle, int bFloat, int nDefault, tMenuItem* menu, int* ascii);
static int DCUIGetSaveFilename(char* pFilename);
static int DCUIGetLoadFilename(char* pDirectory, char* pFilename);
static void DCUIMessage(char* pMessage);
static void DCUIAboutBox();
static void DCUIInit();


tUIDriver dc_ui_driver =
{
	&DCUISelect,
	&DCUIGetSaveFilename,
	&DCUIGetLoadFilename,
	&DCUIMessage, 
	&DCUIAboutBox,
	&DCUIInit
};

static void AboutAtariDC()
{
	static tMenuItem menu_array[] =
	{
		{ "WKEY", ITEM_ENABLED|ITEM_ACTION, NULL, "Press Enter To Continue", NULL, 0 },
		MENU_END
	};

	char* About =
		"AtariDC v" A800DCVERASC " ("__DATE__")\n"
		"by Christian Groessler (c) 2002\n"
		"http://www.groessler.org/a800dc\n"
		"\n"
		"Please report all problems\n"
		"to chris@groessler.org\n"
		"\n"
		"This port is based on\n"
		ATARI_TITLE "\n"
		"http://atari800.atari.org\n"
		"\n"
		"It uses the KallistiOS library\n"
		"http://cadcdev.sourceforge.net\n"
		"\n"
		"\n"
		"\n"
		"\n"
		;
	ui_driver->fSelect(About, FALSE, 0, menu_array, NULL);
}

#ifdef CENTERED_MAIN_MENU

/* provide "centered" names */
static struct repl_names {
	char *sig;
	char *new_name;
} repl_names[] = {
	{ "DISK", "            Disk Management" },
	{ "CART", "          Cartridge Management" },
	{ "XBIN", "         Run BIN file directly" },
	{ "SYST", "             Select System" },
	{ "STER", "           Sound Mono/Stereo" },
	{ "ARTF", "           Artifacting mode" },
	{ "SETT", "            Atari Settings" },
	{ "CONT", "        Back to emulated Atari" },
	{ "REST", "          Reset  (Warm Start)" },
	{ "REBT", "          Reboot (Cold Start)" },
	{ "ABOU", "           About the Emulator" },
	{ "MENU", "                 Menu     " },
	{ "EXIT", "             Exit Emulator" }
};
#define num_repl_names (sizeof(repl_names) / sizeof(struct repl_names))

#endif /* #ifdef CENTERED_MAIN_MENU */

static void EmulatorSettings()
{
	static tMenuItem menu_array[] =
	{
		{ "PADD", ITEM_ENABLED|ITEM_CHECK, NULL, "Emulate paddles:",                  NULL, 0 },
		{ "AFRE", ITEM_ENABLED|ITEM_CHECK, NULL, "Joystick autofire:",                NULL, 1 },
		{ "SNDE", ITEM_ENABLED|ITEM_CHECK, NULL, "Enable sound:",                     NULL, 2 },
		{ "NTSC", ITEM_ENABLED|ITEM_CHECK, NULL, "NTSC video mode:",                  NULL, 3 },
		MENU_END
	};

	int option = 0;

	do {
		if (emulate_paddles)
			menu_array[0].flags |= ITEM_CHECKED;
		else
			menu_array[0].flags &= ~ITEM_CHECKED;

		if (joy_autofire[0])
			menu_array[1].flags |= ITEM_CHECKED;
		else
			menu_array[1].flags &= ~ITEM_CHECKED;

		if (glob_snd_ena)
			menu_array[2].flags |= ITEM_CHECKED;
		else
			menu_array[2].flags &= ~ITEM_CHECKED;

		if (tv_mode == TV_NTSC)
			menu_array[3].flags |= ITEM_CHECKED;
		else
			menu_array[3].flags &= ~ITEM_CHECKED;

		option = ui_driver->fSelect(NULL, TRUE, option, menu_array, NULL);

		switch(option) {
		case 0:
			emulate_paddles = !emulate_paddles;
			break;
		case 1:
			joy_autofire[0] = joy_autofire[0] ? 0 : 1;
			break;
		case 2:
			glob_snd_ena = !glob_snd_ena;
			break;
		case 3:
			if (tv_mode == TV_NTSC) tv_mode = TV_PAL;
			else tv_mode = TV_NTSC;
			break;
		}
	} while(option >= 0);
}


static int DCUISelect(char* pTitle, int bFloat, int nDefault, tMenuItem* menu, int* ascii)
{
	tMenuItem* pNewMenu;
	int i;
#ifdef CENTERED_MAIN_MENU
	int j;
#endif
	int result;
	int itemcount;


        /* Count items in old menu */	
	for(itemcount=0; menu[itemcount].sig; itemcount++) ;
	itemcount ++;

        /* Allocate memory for new menu, one extra item */
	pNewMenu = (tMenuItem*)malloc(sizeof(tMenuItem)*itemcount);

        /* Copy old menu, add/modify items when necessary */
	memcpy(pNewMenu, menu, sizeof(tMenuItem)*itemcount);

        /* Modify old menu items */
	for (i=0; pNewMenu[i].sig; i++) {
		pNewMenu[i].suffix = NULL; /* no keyboard shortcuts in Pocket Atari */

#ifdef CENTERED_MAIN_MENU
		/* modify name (centered version) */
		for (j=0; j<num_repl_names; j++)
			if (strcmp(pNewMenu[i].sig, repl_names[j].sig) == 0) {
				pNewMenu[i].item = repl_names[j].new_name;
				break;
			}
#endif

		if (strcmp(pNewMenu[i].sig, "STER") == 0) {
#ifndef STEREO
                        /* Stereo is disabled based on compile settings */
			pNewMenu[i].flags &= ~ITEM_ENABLED;
#endif
		}
		else if (strcmp(pNewMenu[i].sig, "SREC") == 0) {
                        /* Sound recording is disabled */
			pNewMenu[i].flags &= ~ITEM_ENABLED;
		}
		else if (strcmp(pNewMenu[i].sig, "PCXN") == 0) {
                        /* PCX screenshort is not needed */
			pNewMenu[i].flags &= ~ITEM_ENABLED;
		}
		else if (strcmp(pNewMenu[i].sig, "LOAD") == 0) {
			pNewMenu[i].flags &= ~ITEM_ENABLED;
		}
		else if (strcmp(pNewMenu[i].sig, "SAVE") == 0) {
			pNewMenu[i].flags &= ~ITEM_ENABLED;
		}
		else if (strcmp(pNewMenu[i].sig, "PCXI") == 0) {

                        /* Interlaced PCX replaced by Emulator Settings (old feature was too esoteric) */
	 		pNewMenu[i].sig = "EMUX";
			pNewMenu[i].flags = ITEM_ENABLED|ITEM_SUBMENU;
			pNewMenu[i].prefix = NULL;
			pNewMenu[i].item =
#ifdef CENTERED_MAIN_MENU
				"           "
#endif
				"AtariDC Settings";
			pNewMenu[i].suffix = NULL;
			pNewMenu[i].retval = 1000;
		}
		else if(strcmp(pNewMenu[i].sig, "MONI") == 0) {
			if(strcmp(pNewMenu[0].sig, "DISK") == 0) { /* Main menu? */

                                /* Monitor menu item replaced with About box */
				pNewMenu[i].sig = "ABDC";
				pNewMenu[i].flags = ITEM_ENABLED|ITEM_SUBMENU;
				pNewMenu[i].prefix = NULL;
				pNewMenu[i].item = 
#ifdef CENTERED_MAIN_MENU
					"             "
#endif
					"About AtariDC";
				pNewMenu[i].suffix = NULL;
				pNewMenu[i].retval = 1001;
			}
			else {
				pNewMenu[i].flags &= ~ITEM_ENABLED;
			}
		}
	}
        /* Passthrough to the basic driver */
	result = basic_ui_driver.fSelect(pTitle, bFloat, nDefault, pNewMenu, ascii);

        /* Free memory */
	free((void*)pNewMenu);

        /* done */
	switch(result) {
	case 1000:
		EmulatorSettings();
		break;
	case 1001:
		AboutAtariDC();
		break;
	}
	return result;
}

static int DCUIGetSaveFilename(char* pFilename)
{
	return basic_ui_driver.fGetSaveFilename(pFilename);
}

static int DCUIGetLoadFilename(char* pDirectory, char* pFilename)
{
	return basic_ui_driver.fGetLoadFilename(pDirectory, pFilename);
}

static void DCUIMessage(char* pMessage)
{
	basic_ui_driver.fMessage(pMessage);
}

static void DCUIAboutBox()
{
	basic_ui_driver.fAboutBox();
}

static void DCUIInit()
{
	basic_ui_driver.fInit();
}
